INCLUDE "utils.ml"

open Format
module M = Map.Make(struct type t = int let compare = compare end)

type info = { sat : StateSet.t;
              todo : StateSet.t;
              msg : string;
            }
let info = Hashtbl.create 2017
let final = Hashtbl.create 2017

let max_round = ref 0


let buff = Buffer.create 20
let fmt = formatter_of_buffer buff


let trace ?(msg="") nid r t d  =
  if r > !max_round then max_round := r;
  let m = try Hashtbl.find info nid with Not_found -> M.empty in
  let () = pp_print_flush fmt () in
  let _ = fprintf fmt
    "node: %i<br/>%s<br/>todo: %a<br/>sat: %a<br/>_______________________<br/>"
    nid msg  StateSet.print t StateSet.print d
  in
  let () = pp_print_flush fmt () in
  let msg = Buffer.contents buff in
  let () = Buffer.clear buff in
  let old_inf = try M.find r m with Not_found -> [] in
  let m' = M.add r ({ sat = d; todo = t; msg = msg }::old_inf) m in
  Hashtbl.replace info nid m'

let finalize_node n r b =
  Hashtbl.replace final n (b,r)
module K =
struct
  type t = StateSet.t list
  let hash l =
    List.fold_left (fun acc set ->
      HASHINT2(acc, (set.StateSet.id :> int))) 17 l

  let equal l1 l2 =
    try
      List.for_all2 (==) l1 l2
    with _ -> false
end

module CTable = Hashtbl.Make (K)

let ctable = CTable.create 20

let rgb x =
  let h = K.hash x in
  let r = h land 0xff
  and g = (h lsr 8) land 0xff
  and b = (h lsr 16) land 0xff
  in
  r, g, b

let color x =
  try
    CTable.find ctable x
  with
    Not_found ->
      let r,g,b = rgb x in
      let s = "rgb(" ^ (string_of_int r) ^ ","
        ^ (string_of_int g) ^ ","
        ^ (string_of_int b) ^ ")"
      in
      CTable.add ctable x s;
      s

let text_color x =
  let r,g,b = rgb x in
  let av = (r + g + b) / 3 in
  if av > 128 then "rgb(0,0,0)"
  else "rgb(255,255,255)"

let get_conf sel l i =
  List.fold_left (fun (accb,accl) a ->
    accb || StateSet.intersect a.(i) sel,
    a.(i) :: accl) (false,[]) l

let dup_rev_list l =
  List.fold_left (fun acc e -> e::e::acc) [] l

let gen_trace (type s) = fun auto sat_arrays t tree ->
  let sat_arrays = dup_rev_list sat_arrays in
  let module T = (val (t) : Tree.S with type t = s) in
  let root = T.root tree in
  let sel = Ata.get_selecting_states auto in
  let todos = Ata.get_states_by_rank auto in
  let pr_sets = Pretty.print_list ~sep:", "
    (fun fmt s -> fprintf fmt "'%a'" StateSet.print s)
  in
  let rec loop output node parent prevsib x y =
    if node != T.nil then begin
      let node_id = T.preorder tree node in
      let marked, conf = get_conf sel sat_arrays node_id in
      let scolor, tcolor = color conf, text_color conf in
      let tag = QName.to_string (T.tag tree node) in
      let lbox = (String.length tag + 2) * 10 in
      let s_node = "node" ^ (string_of_int node_id) in
      let first = T.first_child tree node in
      let next = T.next_sibling tree node in
      fprintf output "<script type='text/javascript'><![CDATA[ ";
      let pass, max_active_pass, diff_list, full_list =
        List.fold_left
          (fun (pass, max_active_pass, diff_list, full_list) a ->
            let todo = if pass >= Array.length todos then StateSet.empty
              else todos.(pass)
            in
            let diff = StateSet.inter a.(node_id) todo in
            let new_active_pass =
              if diff == StateSet.empty then max_active_pass else pass
            in
            pass+1, new_active_pass, diff::diff_list, a.(node_id) :: full_list)
          (0, -1, [], []) sat_arrays
      in
      fprintf output "tree['%s'] = { marked: %b, fc: 'node%i', ns: 'node%i', par: 'node%i', ps: 'node%i', max: %i, dlist: [%a], flist: [%a] };]]></script>\n"
        s_node
        marked
        (T.preorder tree first) (T.preorder tree next)
        (T.preorder tree parent) (T.preorder tree prevsib)
        max_active_pass
        pr_sets (List.rev diff_list)
        pr_sets (List.rev full_list);
      fprintf output
        "<rect id=\"%s\" onclick=\"activate(\'%s\');\" x=\"%i\" y=\"%i\"\
 width=\"%i\" height=\"22\" style=\"fill:%s;stroke:rgb(0,0,0)%s\"/>%!"
        s_node
        s_node
        x y
        lbox
        scolor
        (if marked
         then ";stroke-width:4"
         else ";stroke-width:2;stroke-dasharray:2,2");
      fprintf output "<text x=\"%i\" y=\"%i\" style=\"fill:%s;font-size:17;\
font-family:typewriter;\" onclick=\"activate(\'%s\');\" >%s</text>\n"
        (x+10)
        (y+15)
        tcolor s_node tag;
      let maxw1, maxy1 = loop output first node T.nil x (y + 40) in
      let x_next = max (x+lbox) (maxw1+10) in
      if node != root then begin
        if prevsib == T.nil then
          fprintf output  "<line x1=\"%i\" y1=\"%i\" x2=\"%i\" y2=\"%i\"\
style=\"stroke:rgb(0,0,0);stroke-width:2\"/>\n"
            (x + lbox / 2) (y-18) (x + lbox / 2) (y);
        if next != T.nil then
          fprintf output "<line x1=\"%i\" y1=\"%i\" x2=\"%i\" y2=\"%i\"\
style=\"stroke:rgb(0,0,0);stroke-width:2\"/>\n"
            (x + lbox) (y+11) x_next (y+11);
      end;
      let maxw2, maxy2 = loop output next parent node x_next y in
      maxw2, max maxy1 maxy2
    end
    else x, y
  in
  ignore (Sys.command "mkdir -p tests/trace");
  let ohtml_ = open_out "tests/trace/trace.html" in
  let ohtml = formatter_of_out_channel ohtml_ in
  fprintf ohtml "<html>\n\
<head><title></title>\n\
<meta http-equiv='content-type' content='text/html;charset=utf-8'/>\n\
<style>\n\
%s
</style>\n\
</head>\n\
<body>\n\
<div id='automata' >%a</div>\n\
<div id='data' > </div>\n\
<script type='text/javascript'>\n\
var tree = new Array();\n\
var current_node = null;
var nil_id = 'node%i';
var last_id = %i;
var rounds = %i;\n"
    Trace_css.content
    Ata.print auto
    (T.preorder tree T.nil)
    (T.size tree)
    (List.length sat_arrays - 1);
  fprintf ohtml "%s\n" Trace_js.content;
  fprintf ohtml
    "</script>\n\
<div id='tree'>
<div id='controls'><button onclick='activate_parent();'>↑</button>\n\
<button onclick='activate_previous();'>←</button>\n\
<button onclick='activate_first();'>↓</button>\n\
<button onclick='activate_next();'>→</button>\n\
<select id='relnodes' onchange='hide_nodes();'>\n\
<option value='-1'>All nodes</option>\n\
%a</select>\n\
</div>
<div id='svg'>\n
   <svg id='svgimg' width='' height='' xmlns=\"http://www.w3.org/2000/svg\" version=\"1.1\">\n"
    (Pretty.print_range (fun fmt i ->
      fprintf fmt "<option value='%i'>Relevant nodes on pass %i</option>\n" i i))
    (0,(List.length sat_arrays - 1));
  let maxw, maxh = loop ohtml (T.root tree) T.nil T.nil 50 50 in
  fprintf ohtml "</svg>\n</div></div><script type='text/javascript'>\n\
window.onload = function () {\n\
    var svg = document.getElementById('svgimg');\n\
    svg.setAttribute('width', %i);\n\
    svg.setAttribute('height', %i);\n\
    activate('node0');\n\
};\n
</script></body></html>\n%!"
    maxw maxh;
  pp_print_flush ohtml ();
  close_out ohtml_
